from .__about__ import (
    __author__,
    __commit__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __uri__,
    __version__,
)

from . import (
    __main__,
    annotation,
    cli,
    config,
    csv,
    datasets,
    device,
    engine,
    entry_points,
    files,
    io,
    labels,
    labeled_timebins,
    logging,
    metrics,
    models,
    nn,
    plot,
    spect,
    tensorboard,
    timebins,
    transforms,
    split,
    validators,
)

from .engine.model import Model


__all__ = [
    "__main__",
    "annotation",
    "cli",
    "config",
    "csv",
    "datasets",
    "device",
    "engine",
    "entry_points",
    "files",
    "io",
    "labels",
    "labeled_timebins",
    "logging",
    "metrics",
    "Model",
    "models",
    "nn",
    "plot",
    "spect",
    "split",
    "tensorboard",
    "timebins",
    "transforms",
    "validators",
]
